#include <cassert>

#include <algorithm> // 'std::copy_if'...
// #include <execution> // C++17: execution policies...
#include <iostream>
#include <iterator> // 'std::inserter'...
#include <numeric> // C++17: 'std::reduce'...
#include <string>
#include <type_traits>
#include <unordered_map>
#include <vector>

#include "Lambda_expression_.h"

void Illustration_of_mutable::object_state_is_not_changed() const {
    //    _a++; // Compilation error due to 'const'
    _b++; // 'mutable' inhibits 'const'
}

const std::string Election::Candidate = "Candidate";
const std::string Election::Challenger = "Challenger";
const std::string Election::President = "President";
const std::string Election::Prime_minister = "Prime minister";

Election::Election(const std::string& separator, const std::initializer_list<std::unordered_multimap<std::string, std::string>::value_type>& situation) {
    _separator = separator;
    for (std::unordered_multimap<std::string, std::string>::value_type s : situation)
        _situation.insert(s);
}

void Election::change_separator(const std::string& separator) {
    _separator = separator;
}

std::string Election::candidates() const {
    std::string result;

    auto is_candidate_predicate = [](const std::unordered_multimap<std::string, std::string >::value_type & someone_in_role) -> bool {
        // As shown, static members, e.g., 'Election::Candidate', do not require capture:
        return Election::Candidate.compare(someone_in_role.second) == 0;
    };

    //	decltype(is_candidate_predicate) is_candidate_predicate_;
    std::cout << "'std::is_trivially_constructible_v<decltype(is_candidate_predicate)>': " <<
            std::is_trivially_constructible<decltype(is_candidate_predicate)>::value << std::endl; // 'false'

    decltype(is_candidate_predicate) copy_of_is_candidate_predicate = is_candidate_predicate;
    std::cout << "'std::is_trivially_copyable_v<decltype(is_candidate_predicate)>': " <<
            std::is_trivially_copyable<decltype(is_candidate_predicate)>::value << std::endl; // 'true'

    //	copy_of_is_candidate_predicate = is_candidate_predicate;
    std::cout << "'std::is_trivially_assignable_v<decltype(is_candidate_predicate), decltype(is_candidate_predicate)>': " <<
            std::is_trivially_assignable<decltype(is_candidate_predicate), decltype(is_candidate_predicate)>::value << std::endl; // 'false'

    if (std::none_of(std::begin(_situation), std::end(_situation), is_candidate_predicate))
        result += "Nobody is " + Election::Candidate;
    else {
        std::unordered_multimap< std::string, std::string > candidates;
        std::copy_if(/*std::execution::par,*/ _situation.begin(), _situation.end(), std::inserter(candidates, candidates.end()),
                [](auto& someone_in_role) {
                    return Election::Candidate.compare(someone_in_role.second) == 0;
                });

        //        std::copy(_situation.begin(), _situation.end(), std::inserter(candidates, candidates.end())); // Copy all...
        //        'std::remove_if' does not work on maps :
        //                std::remove_if(/*std::execution::par,*/ candidates.begin(), candidates.end(),
        //                [](auto& someone_in_role) {
        //                    return !(Election::Candidate.compare(someone_in_role.second) == 0);
        //                }
        //        );

        std::vector<std::string > keys;
        std::transform(std::begin(candidates), std::end(candidates), std::back_inserter(keys),
                [](decltype(candidates)::value_type & pair) {
                    return pair.first;
                });

        //result = Election::Candidate + ": " + std::reduce(/*std::execution::par,*/ keys.begin(), keys.end(), result,
        //	[this](const std::string& k1, const std::string& k2) {
        //		// Compilation error comes from 'candidates' is a 'const' member function...:
        //		this->change_separator(" # "); // 'change_separator' is a non-'const' member function...
        //		return k1 + this->_separator + k2; // Note that '[=]' is enough to capture 'this->_separator'
        //	});

        //		result = Election::Candidate + ": " + std::reduce(/*std::execution::par,*/ keys.begin(), keys.end(), result,
        //			// Initializer (C++14) and '*this' (C++17):
        //			[my_this = *this](const std::string& k1, const std::string& k2) mutable {
        //			my_this.change_separator(" # "); // 'change_separator' is a non-'const' member function...
        //			return k1 + my_this._separator + k2;
        //		});
    }
    return result;
}

class { // Closure type
public:

    template<typename X> void operator()(X& x) const {

        x++;
    }
} increment;

int main(int argc, char** argv) {
    std::string FB = "Franck Barbier";

    char given_name[] = {'f', 'r', 'a', 'n', 'c', 'k'};
    // Binding requires 6 items:
    auto [f, r, a, n, c, k] = given_name; // Copy 'given_name[i]' into 'f', 'r'...
    assert(f == 'f');

    Person p("Franck", "Barbier", "?");
    auto& [gn, sn, nn] = p; // References to p's (public only!) data...
    assert(gn.compare("Franck") == 0);
    assert(sn.compare("Barbier") == 0);
    nn = "Bab";
    std::cout << p._nickname << std::endl; // 'Bab'
    static_assert(std::is_same<decltype(nn), std::string>::value);

    char my_char = 'a', number_of_my_char = 0;
    // Injecting one object by copy and the other one by reference:
    std::for_each(FB.begin(), FB.end(), [my_char, &number_of_my_char](const char& c) /* -> bool */ {
        if (c == my_char) {
            number_of_my_char++;
            return true;
        }
        return false;
    });
    assert(number_of_my_char == 2); // Two occurrences of 'a'

    char from = 'a', to = 'A';
    bool changed = false;
    auto my_mutable_functor = [from, to, changed](char& c) mutable {
        if (c == from) {
            c = to;
            changed = !changed; // Requires 'mutable'...
        }
        return changed;
    };
    std::for_each(FB.begin(), FB.end(), my_mutable_functor);
    std::cout << FB << std::endl; // As expected: 'FrAnck BArbier'

    from = 'b';
    to = 'B'; // Pay attention: capture does not mean argument passing!
    std::for_each(FB.begin(), FB.end(), my_mutable_functor);
    std::cout << FB << std::endl; // No effect: 'FrAnck BArbier'

    int x = 0;
    // More simply, '::increment(x);'
    ::increment.operator()/*<int>*/(x); // Instantiation of template member function in closure type
    std::cout << "::increment(x): " << x << std::endl; // '1'
    auto increment = [](auto& x) { // Generic lambda...
        x++;
    };
    increment(x); // Instantiation of generic lambda...
    std::cout << "increment(x): " << x << std::endl; // '2'

    std::cout << std::boolalpha << std::is_same<std::pair<const std::string, std::string>,
            std::unordered_map<std::string, std::string >::value_type >::value << std::endl; // 'true'

    std::unordered_multimap<std::string, std::string> situation_of_French_2012_presidential_election{
        {"Macron", "President"},
        {"Macron", "Candidate"},
        {"Philippe", "Prime minister"},
        {"Mélanchon", "Challenger"},
        {"Mélanchon", "Candidate"},
        {"Le Pen", "Challenger"},
        {"Le Pen", "Candidate"}};

    for (auto&& [name, role] : situation_of_French_2012_presidential_election) { // Binding (C++17)
        std::cout << name << " as " << role << std::endl;
    }

    std::unordered_multimap<std::string, std::string >::iterator iterator = situation_of_French_2012_presidential_election.begin();
    increment(iterator); // Instantiation of generic lambda...
    std::cout << "increment(iterator): " << (*iterator).second << std::endl; // No order, so... value of any element

    Election French_2012_presidential_election{ " - ", // 'std::runtime_error'
        {
            {"Macron", Election::President},
            {"Macron", Election::Candidate},
            {"Philippe", Election::Prime_minister},
            {"Mélanchon", Election::Challenger},
            {"Mélanchon", Election::Candidate},
            {"Le Pen", Election::Challenger},
            {"Le Pen", Election::Candidate}
        }};

    std::cout << French_2012_presidential_election.candidates() << std::endl;

    return 0;
}

